package com.oracle.webservices.internal.api.message;

import java.lang.reflect.InvocationTargetException;
import com.sun.istack.internal.Nullable;
import com.sun.istack.internal.NotNull;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.AbstractMap;
import java.util.HashSet;
import java.security.AccessController;
import java.lang.reflect.Method;
import java.lang.reflect.Field;
import java.security.PrivilegedAction;
import java.util.Map;

public abstract class BasePropertySet implements PropertySet
{
    private Map<String, Object> mapView;
    
    protected BasePropertySet() {
    }
    
    protected abstract BasePropertySet.PropertyMap getPropertyMap();
    
    protected static BasePropertySet.PropertyMap parse(final Class clazz) {
        return (BasePropertySet.PropertyMap)AccessController.doPrivileged(new PrivilegedAction<BasePropertySet.PropertyMap>() {
            @Override
            public BasePropertySet.PropertyMap run() {
                BasePropertySet.PropertyMap props = new BasePropertySet.PropertyMap();
                for (Class c = clazz; c != null; c = c.getSuperclass()) {
                    for (Field f : c.getDeclaredFields()) {
                        PropertySet.Property cp = (PropertySet.Property)f.getAnnotation(PropertySet.Property.class);
                        if (cp != null) {
                            for (String value : cp.value()) {
                                props.put(value, new BasePropertySet.FieldAccessor(f, value));
                            }
                        }
                    }
                    for (Method m : c.getDeclaredMethods()) {
                        PropertySet.Property cp = (PropertySet.Property)m.getAnnotation(PropertySet.Property.class);
                        if (cp != null) {
                            String name = m.getName();
                            assert !(!name.startsWith("is"));
                            String setName = name.startsWith("is") ? ("set" + name.substring(2)) : (115 + name.substring(1));
                            Method setter;
                            try {
                                setter = clazz.getMethod(setName, new Class[] { m.getReturnType() });
                            }
                            catch (NoSuchMethodException e) {
                                setter = null;
                            }
                            for (String value2 : cp.value()) {
                                props.put(value2, new BasePropertySet.MethodAccessor(m, setter, value2));
                            }
                        }
                    }
                }
                return props;
            }
        });
    }
    
    @Override
    public boolean containsKey(Object key) {
        BasePropertySet.Accessor sp = (BasePropertySet.Accessor)this.getPropertyMap().get(key);
        return sp != null && sp.get(this) != null;
    }
    
    @Override
    public Object get(Object key) {
        BasePropertySet.Accessor sp = (BasePropertySet.Accessor)this.getPropertyMap().get(key);
        if (sp != null) {
            return sp.get(this);
        }
        throw new IllegalArgumentException("Undefined property " + key);
    }
    
    @Override
    public Object put(String key, Object value) {
        BasePropertySet.Accessor sp = (BasePropertySet.Accessor)this.getPropertyMap().get(key);
        if (sp != null) {
            Object old = sp.get(this);
            sp.set(this, value);
            return old;
        }
        throw new IllegalArgumentException("Undefined property " + key);
    }
    
    @Override
    public boolean supports(Object key) {
        return this.getPropertyMap().containsKey(key);
    }
    
    @Override
    public Object remove(Object key) {
        BasePropertySet.Accessor sp = (BasePropertySet.Accessor)this.getPropertyMap().get(key);
        if (sp != null) {
            Object old = sp.get(this);
            sp.set(this, null);
            return old;
        }
        throw new IllegalArgumentException("Undefined property " + key);
    }
    
    @Deprecated
    @Override
    public final Map<String, Object> createMapView() {
        final Set<Map.Entry<String, Object>> core = new HashSet();
        this.createEntrySet(core);
        return new AbstractMap<String, Object>() {
            @Override
            public Set<Map.Entry<String, Object>> entrySet() {
                return core;
            }
        };
    }
    
    @Override
    public Map<String, Object> asMap() {
        if (this.mapView == null) {
            this.mapView = this.createView();
        }
        return this.mapView;
    }
    
    protected Map<String, Object> createView() {
        return new BasePropertySet.MapView(this.mapAllowsAdditionalProperties());
    }
    
    protected boolean mapAllowsAdditionalProperties() {
        return false;
    }
    
    protected void createEntrySet(Set<Map.Entry<String, Object>> core) {
        for (Map.Entry<String, BasePropertySet.Accessor> e : this.getPropertyMap().entrySet()) {
            core.add(new Map.Entry<String, Object>() {
                @Override
                public String getKey() {
                    return (String)e.getKey();
                }
                
                @Override
                public Object getValue() {
                    return ((BasePropertySet.Accessor)e.getValue()).get(BasePropertySet.this);
                }
                
                @Override
                public Object setValue(Object value) {
                    BasePropertySet.Accessor acc = (BasePropertySet.Accessor)e.getValue();
                    Object old = acc.get(BasePropertySet.this);
                    acc.set(BasePropertySet.this, value);
                    return old;
                }
            });
        }
    }
    
    protected static class PropertyMap extends HashMap<String, BasePropertySet.Accessor>
    {
        transient BasePropertySet.PropertyMapEntry[] cachedEntries;
        
        protected PropertyMap() {
            this.cachedEntries = null;
        }
        
        BasePropertySet.PropertyMapEntry[] getPropertyMapEntries() {
            if (this.cachedEntries == null) {
                this.cachedEntries = this.createPropertyMapEntries();
            }
            return this.cachedEntries;
        }
        
        private BasePropertySet.PropertyMapEntry[] createPropertyMapEntries() {
            BasePropertySet.PropertyMapEntry[] modelEntries = new BasePropertySet.PropertyMapEntry[this.size()];
            int i = 0;
            for (Map.Entry<String, BasePropertySet.Accessor> e : this.entrySet()) {
                modelEntries[i++] = new BasePropertySet.PropertyMapEntry((String)e.getKey(), (BasePropertySet.Accessor)e.getValue());
            }
            return modelEntries;
        }
    }
    
    public static class PropertyMapEntry
    {
        String key;
        BasePropertySet.Accessor value;
        
        public PropertyMapEntry(String k, BasePropertySet.Accessor v) {
            this.key = k;
            this.value = v;
        }
    }
    
    static final class FieldAccessor implements BasePropertySet.Accessor
    {
        private final Field f;
        private final String name;
        
        protected FieldAccessor(Field f, String name) {
            (this.f = f).setAccessible(true);
            this.name = name;
        }
        
        @Override
        public String getName() {
            return this.name;
        }
        
        @Override
        public boolean hasValue(PropertySet props) {
            return this.get(props) != null;
        }
        
        @Override
        public Object get(PropertySet props) {
            try {
                return this.f.get(props);
            }
            catch (IllegalAccessException e) {
                throw new AssertionError();
            }
        }
        
        @Override
        public void set(PropertySet props, Object value) {
            try {
                this.f.set(props, value);
            }
            catch (IllegalAccessException e) {
                throw new AssertionError();
            }
        }
    }
    
    static final class MethodAccessor implements BasePropertySet.Accessor
    {
        @NotNull
        private final Method getter;
        @Nullable
        private final Method setter;
        private final String name;
        
        protected MethodAccessor(Method getter, Method setter, String value) {
            this.getter = getter;
            this.setter = setter;
            this.name = value;
            getter.setAccessible(true);
            if (setter != null) {
                setter.setAccessible(true);
            }
        }
        
        @Override
        public String getName() {
            return this.name;
        }
        
        @Override
        public boolean hasValue(PropertySet props) {
            return this.get(props) != null;
        }
        
        @Override
        public Object get(PropertySet props) {
            try {
                return this.getter.invoke(props, new Object[0]);
            }
            catch (IllegalAccessException e) {
                throw new AssertionError();
            }
            catch (InvocationTargetException e2) {
                this.handle(e2);
                return Integer.valueOf(0);
            }
        }
        
        @Override
        public void set(PropertySet props, Object value) {
            if (this.setter == null) {
                throw new ReadOnlyPropertyException(this.getName());
            }
            try {
                this.setter.invoke(props, new Object[] { value });
            }
            catch (IllegalAccessException e) {
                throw new AssertionError();
            }
            catch (InvocationTargetException e2) {
                this.handle(e2);
            }
            return;
        }
        
        private Exception handle(InvocationTargetException e) {
            Throwable t = e.getTargetException();
            if ((t instanceof Error)) {
                throw (Error)t;
            }
            if ((t instanceof RuntimeException)) {
                throw (RuntimeException)t;
            }
            throw new Error(e);
        }
    }
    
    final class MapView extends HashMap<String, Object>
    {
        boolean extensible;
        
        MapView(boolean extensible) {
            super(BasePropertySet.this.getPropertyMap().getPropertyMapEntries().length);
            this.extensible = extensible;
            this.initialize();
        }
        
        public void initialize() {
            BasePropertySet.PropertyMapEntry[] entries;
            for (BasePropertySet.PropertyMapEntry entry : entries = BasePropertySet.this.getPropertyMap().getPropertyMapEntries()) {
                super.put(entry.key, entry.value);
            }
        }
        
        @Override
        public Object get(Object key) {
            Object o = super.get(key);
            if ((o instanceof BasePropertySet.Accessor)) {
                return ((BasePropertySet.Accessor)o).get(BasePropertySet.this);
            }
            return o;
        }
        
        @Override
        public Set<Map.Entry<String, Object>> entrySet() {
            Set<Map.Entry<String, Object>> entries = new HashSet();
            for (String key : this.keySet()) {
                entries.add(new AbstractMap.SimpleImmutableEntry(key, this.get(key)));
            }
            return entries;
        }
        
        @Override
        public Object put(String key, Object value) {
            Object o = super.get(key);
            if (o != null && o instanceof BasePropertySet.Accessor) {
                Object oldValue = ((BasePropertySet.Accessor)o).get(BasePropertySet.this);
                ((BasePropertySet.Accessor)o).set(BasePropertySet.this, value);
                return oldValue;
            }
            if (this.extensible) {
                return super.put(key, value);
            }
            throw new IllegalStateException("Unknown property [" + key + "] for PropertySet [" + BasePropertySet.this.getClass().getName() + "]");
        }
        
        @Override
        public void clear() {
            for (String key : this.keySet()) {
                this.remove(key);
            }
        }
        
        @Override
        public Object remove(Object key) {
            Object o = super.get(key);
            if ((o instanceof BasePropertySet.Accessor)) {
                ((BasePropertySet.Accessor)o).set(BasePropertySet.this, null);
            }
            return super.remove(key);
        }
    }
    
    protected interface Accessor
    {
        public abstract String getName();
        
        public abstract boolean hasValue(PropertySet p0);
        
        public abstract Object get(PropertySet p0);
        
        public abstract void set(PropertySet p0, Object p1);
    }
}
